from __future__ import annotations

__all__ = ["MNIST", "get_default_transform"]

from collections.abc import Callable
from pathlib import Path
from unittest.mock import Mock

from gravitorch import constants as ct
from gravitorch.utils.factory import setup_object
from gravitorch.utils.imports import check_torchvision, is_torchvision_available
from gravitorch.utils.path import sanitize_path

if is_torchvision_available():
    from torchvision import transforms
    from torchvision.datasets import MNIST as MNIST_
    from torchvision.transforms import Compose
else:
    MNIST_, Compose = Mock, "Compose"  # pragma: no cover


class MNIST(MNIST_):
    r"""Updated MNIST dataset to return a dict instead of a tuple."""

    def __init__(
        self,
        root: Path | str,
        train: bool = True,
        transform: Callable | dict | None = None,
        target_transform: Callable | dict | None = None,
        download: bool = False,
    ) -> None:
        check_torchvision()
        super().__init__(
            root=sanitize_path(root).as_posix(),
            train=train,
            transform=setup_object(transform),
            target_transform=setup_object(target_transform),
            download=download,
        )

    def __getitem__(self, index: int) -> dict:
        r"""Get the image and the target of the index-th example.

        Args:
        ----
            index (int): Specifies the index of the example.

        Returns:
        -------
            dict: A dictionary with the image and the target of the
                ``index``-th example.
        """
        img, target = super().__getitem__(index)
        return {ct.INPUT: img, ct.TARGET: target}

    @classmethod
    def create_with_default_transforms(
        cls,
        root: Path | str,
        train: bool = True,
        download: bool = False,
    ) -> MNIST:
        r"""Creates a MNIST dataset wih default transforms.

        Args:
            root (``Path`` or str): Specifies the root directory of
                the dataset.
            train (bool, optional): If ``True``, returns the training
                dataset, otherwise returns the test dataset.
                Default: ``True``
            download (bool, optional): If ``True``, downloads the
                dataset from the internet and puts it in root
                directory. If dataset is already downloaded, it is
                not downloaded again. Default: ``False``

        Returns:
            ``MNIST``: A MNIST dataset with default transforms.
        """
        return cls(
            root=root,
            train=train,
            download=download,
            transform=get_default_transform(),
            target_transform=None,
        )


def get_default_transform() -> Compose:
    r"""Gets the default transform for MNIST dataset.

    Returns:
        ``torchvision.transforms.Compose``: The default transforms.

    Raises:
        RuntimeError if ``torchvision`` is not installed.
    """
    check_torchvision()
    return Compose([transforms.ToTensor(), transforms.Normalize((0.1307,), (0.3081,))])



r"""This module implements a data source for MNIST dataset."""

from __future__ import annotations

__all__ = ["MnistDataSource"]

from pathlib import Path

from torch.utils.data import Dataset

from gravitorch import constants as ct
from gravitorch.creators.dataloader.base import BaseDataLoaderCreator
from gravitorch.data.datasets import MNIST
from gravitorch.data.datasets.mnist import get_default_transform
from gravitorch.datasources.dataset import DatasetDataSource
from gravitorch.utils.path import sanitize_path


class MnistDataSource(DatasetDataSource):
    r"""Implements a data source for the MNIST dataset.

    Args:
    ----
        path (str, optional): Specifies the path where to save/load
            the MNIST data.
        data_loader_creators (dict): Specifies the data loader
            creators to initialize. Each key indicates a data loader
            creator name. The value can be a ``BaseDataLoaderCreator``
            object, or its configuration, or ``None``. ``None`` means
            a default data loader will be created. Each data loader
            creator takes a ``Dataset`` object as input, so you need
            to specify a dataset with the same name. The expected
            keys are ``'train'`` and ``'eval'``.
    """

    def __init__(
        self,
        path: Path | str,
        data_loader_creators: dict[str, BaseDataLoaderCreator | dict | None],
    ) -> None:
        train_dataset, eval_dataset = create_standard_train_eval_dataset(sanitize_path(path))
        super().__init__(
            datasets={ct.TRAIN: train_dataset, ct.EVAL: eval_dataset},
            data_loader_creators=data_loader_creators,
        )


def create_standard_train_eval_dataset(path: Path) -> tuple[Dataset, Dataset]:
    r"""Creates the standard training and evaluation MNIST datasets.

    Args:
    ----
        path (``pathlib.Path``): Specifies the path where to save/load
            MNIST data.

    Returns:
    -------
        ``Dataset, Dataset``: The training (first value in the tuple)
            and evaluation (second value in the tuple) datasets.
    """
    transform = get_default_transform()
    return (
        MNIST(path.as_posix(), train=True, download=True, transform=transform),
        MNIST(path.as_posix(), train=False, download=True, transform=transform),
    )



r"""The data package contains the data loader base class and some tools
to speed up the implementation or setup of new data loaders."""

from __future__ import annotations

__all__ = ["create_dataloader", "create_dataloader2"]

from gravitorch.data.dataloaders.factory import create_dataloader, create_dataloader2



from __future__ import annotations

__all__ = ["IterableDataFlow"]

from collections.abc import Iterable, Iterator

from gravitorch.experimental.dataflow.base import BaseDataFlow


class IterableDataFlow(BaseDataFlow):
    r"""Implements a simple dataflow for iterables.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.experimental.dataflow import IterableDataFlow
        >>> with IterableDataFlow([1, 2, 3, 4, 5]) as dataflow:
        ...     for batch in dataflow:
        ...         print(batch)  # do something
        ...
    """

    def __init__(self, iterable: Iterable) -> None:
        if not isinstance(iterable, Iterable):
            raise TypeError(f"Incorrect type. Expecting iterable but received {type(iterable)}")
        self.iterable = iterable

    def __iter__(self) -> Iterator:
        return iter(self.iterable)

    def __str__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def launch(self) -> None:
        r"""Nothing to do for this dataflow."""

    def shutdown(self) -> None:
        r"""Nothing to do for this dataflow."""



from __future__ import annotations

__all__ = ["create_dataloader", "create_dataloader2"]

from collections.abc import Iterable

from torch.utils.data import DataLoader, Dataset
from torch.utils.data.graph import DataPipe

from gravitorch.data.datasets.factory import setup_dataset
from gravitorch.utils.factory import setup_object
from gravitorch.utils.imports import check_torchdata, is_torchdata_available

if is_torchdata_available():
    from torchdata.dataloader2 import DataLoader2, ReadingServiceInterface
    from torchdata.dataloader2.adapter import Adapter
else:  # pragma: no cover
    Adapter = "Adapter"
    DataLoader2 = "DataLoader2"
    ReadingServiceInterface = "ReadingServiceInterface"


def create_dataloader(dataset: Dataset | dict, **kwargs) -> DataLoader:
    r"""Instantiates a ``torch.utils.data.DataLoader`` from a
    ``torch.utils.data.Dataset`` or its configuration.

    Args:
        dataset (``torch.utils.data.Dataset`` or ``dict``): Specifies
            a dataset or its configuration.
        **kwargs: See ``torch.utils.data.DataLoader`` documentation.

    Returns:
        ``torch.utils.data.DataLoader``: The instantiated dataloader.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.data.dataloaders import create_dataloader
        >>> create_dataloader(
        ...     {
        ...         "_target_": "gravitorch.data.datasets.DummyMultiClassDataset",
        ...         "num_examples": 10,
        ...         "num_classes": 2,
        ...         "feature_size": 4,
        ...     }
        ... )
        <torch.utils.data.dataloader.DataLoader at 0x0123456789>
    """
    return DataLoader(
        setup_dataset(dataset), **{key: setup_object(value) for key, value in kwargs.items()}
    )


def create_dataloader2(
    datapipe: DataPipe | dict,
    datapipe_adapter_fn: Iterable[Adapter | dict] | Adapter | dict | None = None,
    reading_service: ReadingServiceInterface | dict | None = None,
) -> DataLoader2:
    r"""Instantiates a ``torchdata.dataloader2.DataLoader2`` object
    from a ``DataPipe`` or its configuration.

    Args:
        datapipe: Specifies the DataPipe or its configuration.
        datapipe_adapter_fn: Specifies the ``Adapter`` function(s) that
            will be applied to the DataPipe. Default: ``None``
        reading_service: Defines how ``DataLoader2`` should execute
            operations over the ``DataPipe``, e.g.
            multiprocessing/distributed. Default: ``None``

    Returns:
        ``torchdata.dataloader2.DataLoader2``: The instantiated
            ``DataLoader2``

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.data.dataloaders import create_dataloader2
        >>> create_dataloader2(
        ...     {
        ...         "_target_": "torch.utils.data.datapipes.iter.IterableWrapper",
        ...         "iterable": range(10),
        ...     }
        ... )
        <torchdata.dataloader2.dataloader2.DataLoader2 at 0x0123456789>
    """
    check_torchdata()
    return DataLoader2(
        datapipe=setup_object(datapipe),
        datapipe_adapter_fn=setup_object(datapipe_adapter_fn),
        reading_service=setup_object(reading_service),
    )
